from django.utils.translation import gettext_lazy as _

import logging
from yubico_client import Yubico
from yubico_client.otp import OTP
from yubico_client.yubico_exceptions import YubicoError

from trench.backends.base import AbstractMessageDispatcher
from trench.responses import DispatchResponse, SuccessfulDispatchResponse
from trench.settings import YUBICLOUD_CLIENT_ID


class YubiKeyMessageDispatcher(AbstractMessageDispatcher):
    def dispatch_message(self) -> DispatchResponse:
        return SuccessfulDispatchResponse(details=_("Generate code using YubiKey"))

    def confirm_activation(self, code: str) -> None:
        self._mfa_method.secret = OTP(code).device_id
        self._mfa_method.save(update_fields=("secret",))

    def validate_confirmation_code(self, code) -> bool:
        """
        The `device_id` is not stored in db yet.
        After successful confirmation `validate_code` should be used.
        """
        return self._validate_yubikey_otp(code)

    def validate_code(self, code: str) -> bool:
        if (
            not self._mfa_method.secret
            or self._mfa_method.secret != OTP(code).device_id
        ):
            return False
        return self._validate_yubikey_otp(code)

    def _validate_yubikey_otp(self, code: str) -> bool:
        try:
            return Yubico(self._config[YUBICLOUD_CLIENT_ID]).verify(
                code, timestamp=True
            )
        except (YubicoError, Exception) as cause:
            logging.error(cause, exc_info=True)
            return False



from django.db.models import Model

from abc import ABC, abstractmethod
from pyotp import TOTP
from typing import Any, Dict, Optional, Tuple

from trench.command.create_otp import create_otp_command
from trench.exceptions import MissingConfigurationError
from trench.models import MFAMethod
from trench.responses import DispatchResponse
from trench.settings import SOURCE_FIELD, VALIDITY_PERIOD, trench_settings


class AbstractMessageDispatcher(ABC):
    def __init__(self, mfa_method: MFAMethod, config: Dict[str, Any]) -> None:
        self._mfa_method = mfa_method
        self._config = config
        self._to = self._get_source_field()

    def _get_source_field(self) -> Optional[str]:
        if SOURCE_FIELD in self._config:
            source = self._get_nested_attr_value(
                self._mfa_method.user, self._config[SOURCE_FIELD]
            )
            if source is None:
                raise MissingConfigurationError(
                    attribute_name=self._config[SOURCE_FIELD]
                )
            return source
        return None

    def _get_nested_attr_value(self, obj: Model, path: str) -> Optional[str]:
        objects, attr = self._parse_dotted_path(path)
        try:
            _obj = self._get_innermost_object(obj, objects)
        except AttributeError:  # pragma: no cover
            return None  # pragma: no cover
        return getattr(_obj, attr)

    @staticmethod
    def _parse_dotted_path(path: str) -> Tuple[Optional[str], str]:
        """
        Extracts attribute name from dotted path.
        """
        try:
            objects, attr = path.rsplit(".", 1)
            return objects, attr
        except ValueError:
            return None, path

    @staticmethod
    def _get_innermost_object(obj: Model, dotted_path: str = None) -> Model:
        """
        For given object return innermost object.
        """
        if dotted_path is None:
            return obj
        for o in dotted_path.split("."):
            obj = getattr(obj, o)
        return obj  # pragma: no cover

    @abstractmethod
    def dispatch_message(self) -> DispatchResponse:
        raise NotImplementedError  # pragma: no cover

    def create_code(self) -> str:
        return self._get_otp().now()

    def confirm_activation(self, code: str) -> None:
        pass

    def validate_confirmation_code(self, code: str) -> bool:
        return self.validate_code(code)

    def validate_code(self, code: str) -> bool:
        return self._get_otp().verify(otp=code)

    def _get_otp(self) -> TOTP:
        return create_otp_command(
            secret=self._mfa_method.secret,
            interval=self._get_valid_window()
        )

    def _get_valid_window(self) -> int:
        return self._config.get(
            VALIDITY_PERIOD, trench_settings.DEFAULT_VALIDITY_PERIOD
        )



from django.urls import path

from trench.views.jwt import MFAFirstStepJWTView, MFASecondStepJWTView


urlpatterns = (
    path("login/", MFAFirstStepJWTView.as_view(), name="generate-code-jwt"),
    path("login/code/", MFASecondStepJWTView.as_view(), name="generate-token-jwt"),
)



from django.contrib.auth.models import User

import logging

from trench.backends.base import AbstractMessageDispatcher
from trench.responses import (
    DispatchResponse,
    FailedDispatchResponse,
    SuccessfulDispatchResponse,
)
from trench.settings import trench_settings


class ApplicationMessageDispatcher(AbstractMessageDispatcher):
    def dispatch_message(self) -> DispatchResponse:
        try:
            qr_link = self._create_qr_link(self._mfa_method.user)
            return SuccessfulDispatchResponse(details=qr_link)
        except Exception as cause:  # pragma: nocover
            logging.error(cause, exc_info=True)  # pragma: nocover
            return FailedDispatchResponse(details=str(cause))  # pragma: nocover

    def _create_qr_link(self, user: User) -> str:
        return self._get_otp().provisioning_uri(
            getattr(user, User.USERNAME_FIELD),
            trench_settings.APPLICATION_ISSUER_NAME,
        )



from django.contrib.auth.models import User
from django.db.models import Model

from abc import abstractmethod
from rest_framework.authtoken.models import Token
from rest_framework.fields import CharField, ChoiceField
from rest_framework.serializers import ModelSerializer, Serializer
from typing import Any, Iterable, OrderedDict, Type

from trench.backends.provider import get_mfa_handler
from trench.command.remove_backup_code import remove_backup_code_command
from trench.command.validate_backup_code import validate_backup_code_command
from trench.exceptions import (
    CodeInvalidOrExpiredError,
    MFAMethodAlreadyActiveError,
    MFAMethodDoesNotExistError,
    MFANotEnabledError,
    OTPCodeMissingError,
)
from trench.models import MFAMethod
from trench.settings import trench_settings
from trench.utils import available_method_choices, get_mfa_model


def generate_model_serializer(name: str, model: Model, fields: Iterable[str]) -> Type:
    meta_subclass = type(
        "Meta",
        (object,),
        {
            "model": model,
            "fields": fields,
        },
    )
    return type(name, (ModelSerializer,), {"Meta": meta_subclass})


class RequestBodyValidator(Serializer):
    def update(self, instance: Model, validated_data: OrderedDict[str, Any]):
        raise NotImplementedError

    def create(self, validated_data: OrderedDict[str, Any]):
        raise NotImplementedError


class ProtectedActionValidator(RequestBodyValidator):
    code = CharField()

    @staticmethod
    def _get_validation_method_name() -> str:
        return "validate_code"

    @staticmethod
    @abstractmethod
    def _validate_mfa_method(mfa: MFAMethod) -> None:
        raise NotImplementedError

    def __init__(self, mfa_method_name: str, user: User, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._user = user
        self._mfa_method_name = mfa_method_name

    def validate_code(self, value: str) -> str:
        if not value:
            raise OTPCodeMissingError()
        mfa_model = get_mfa_model()
        mfa = mfa_model.objects.get_by_name(
            user_id=self._user.id, name=self._mfa_method_name
        )
        self._validate_mfa_method(mfa)

        validated_backup_code = validate_backup_code_command(
            value=value, backup_codes=mfa.backup_codes
        )

        handler = get_mfa_handler(mfa)
        validation_method = getattr(handler, self._get_validation_method_name())
        if validation_method(value):
            return value

        if validated_backup_code:
            remove_backup_code_command(
                user_id=mfa.user_id, method_name=mfa.name, code=value
            )
            return value

        raise CodeInvalidOrExpiredError()


class MFAMethodDeactivationValidator(ProtectedActionValidator):
    code = CharField(required=trench_settings.CONFIRM_DISABLE_WITH_CODE)

    @staticmethod
    def _validate_mfa_method(mfa: MFAMethod) -> None:
        if not mfa.is_active:
            raise MFANotEnabledError()


class MFAMethodActivationConfirmationValidator(ProtectedActionValidator):
    @staticmethod
    def _get_validation_method_name() -> str:
        return "validate_confirmation_code"

    @staticmethod
    def _validate_mfa_method(mfa: MFAMethod) -> None:
        if mfa.is_active:
            raise MFAMethodAlreadyActiveError()


class MFAMethodBackupCodesGenerationValidator(ProtectedActionValidator):
    code = CharField(
        required=trench_settings.CONFIRM_BACKUP_CODES_REGENERATION_WITH_CODE
    )

    @staticmethod
    def _validate_mfa_method(mfa: MFAMethod) -> None:
        if not mfa.is_active:
            raise MFANotEnabledError()


class MFAMethodCodeSerializer(RequestBodyValidator):
    method = CharField(max_length=255, required=False)

    @staticmethod
    def validate_method(value: str) -> str:
        if value and value not in trench_settings.MFA_METHODS:
            raise MFAMethodDoesNotExistError()
        return value


class LoginSerializer(RequestBodyValidator):
    password = CharField(write_only=True)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.fields[User.USERNAME_FIELD] = CharField()


class CodeLoginSerializer(RequestBodyValidator):
    ephemeral_token = CharField()
    code = CharField()


class UserMFAMethodSerializer(ModelSerializer):
    class Meta:
        model = get_mfa_model()
        fields = ("name", "is_primary")


class ChangePrimaryMethodValidator(ProtectedActionValidator):
    method = ChoiceField(choices=available_method_choices())

    @staticmethod
    def _validate_mfa_method(mfa: MFAMethod) -> None:
        pass


class TokenSerializer(ModelSerializer):
    auth_token = CharField(source="key")

    class Meta:
        model = Token
        fields = ("auth_token",)



from django.db.transaction import atomic

from typing import Type

from trench.exceptions import DeactivationOfPrimaryMFAMethodError, MFANotEnabledError
from trench.models import MFAMethod
from trench.utils import get_mfa_model


class DeactivateMFAMethodCommand:
    def __init__(self, mfa_model: Type[MFAMethod]) -> None:
        self._mfa_model = mfa_model

    @atomic
    def execute(self, mfa_method_name: str, user_id: int) -> None:
        mfa = self._mfa_model.objects.get_by_name(user_id=user_id, name=mfa_method_name)
        number_of_active_mfa_methods = self._mfa_model.objects.filter(
            user_id=user_id, is_active=True
        ).count()
        if mfa.is_primary and number_of_active_mfa_methods > 1:
            raise DeactivationOfPrimaryMFAMethodError()
        if not mfa.is_active:
            raise MFANotEnabledError()

        self._mfa_model.objects.filter(user_id=user_id, name=mfa_method_name).update(
            is_active=False
        )


deactivate_mfa_method_command = DeactivateMFAMethodCommand(
    mfa_model=get_mfa_model()
).execute



from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

from rest_framework.serializers import ValidationError


class MissingConfigurationError(ImproperlyConfigured):
    def __init__(self, attribute_name: str) -> None:
        super().__init__(f"Could not retrieve attribute '{attribute_name}'.")


class MethodHandlerMissingError(ImproperlyConfigured):
    def __init__(self, method_name: str) -> None:
        super().__init__(f"Missing handler in {method_name} configuration.")


class MFAValidationError(ValidationError):
    def __str__(self) -> str:
        return ", ".join(detail for detail in self.detail)


class CodeInvalidOrExpiredError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(
            detail=_("Code invalid or expired."),
            code="code_invalid_or_expired",
        )


class OTPCodeMissingError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(detail=_("OTP code not provided."), code="otp_code_missing")


class MFAMethodDoesNotExistError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(
            detail=_("Requested MFA method does not exist."),
            code="mfa_method_does_not_exist",
        )


class MFAPrimaryMethodInactiveError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(
            detail=_("MFA Method selected as new primary method is not active"),
            code="new_primary_method_inactive",
        )


class MFAMethodAlreadyActiveError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(
            detail=_("MFA method already active."),
            code="method_already_active",
        )


class DeactivationOfPrimaryMFAMethodError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(
            detail=_(
                "Deactivation of MFA method that is set as primary is not allowed."
            ),
            code="deactivation_of_primary",
        )


class MFANotEnabledError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(detail=_("2FA is not enabled."), code="not_enabled")


class InvalidTokenError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(detail=_("Invalid or expired token."), code="invalid_token")


class InvalidCodeError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(detail=_("Invalid or expired code."), code="invalid_code")


class UnauthenticatedError(MFAValidationError):
    def __init__(self) -> None:
        super().__init__(
            detail=_("Unable to login with provided credentials."),
            code="unauthenticated",
        )



from django.conf import settings
from django.utils.translation import gettext_lazy as _

import string
from rest_framework.settings import APISettings, perform_import
from typing import Any, Dict

from trench.exceptions import MethodHandlerMissingError


class TrenchAPISettings(APISettings):
    _FIELD_USER_SETTINGS = "_user_settings"
    _FIELD_TRENCH_AUTH = "TRENCH_AUTH"
    _FIELD_BACKUP_CODES_CHARACTERS = "BACKUP_CODES_CHARACTERS"
    _FIELD_MFA_METHODS = "MFA_METHODS"
    _FIELD_HANDLER = "HANDLER"

    @property
    def user_settings(self) -> Dict[str, Any]:
        if not hasattr(self, self._FIELD_USER_SETTINGS):
            self._user_settings = getattr(settings, self._FIELD_TRENCH_AUTH, {})
        return self._user_settings

    def __getattr__(self, attr: str) -> Any:
        val = super().__getattr__(attr)
        self._validate(attribute=attr, value=val)
        return val

    def _validate(self, attribute: str, value: Any) -> None:
        if attribute == self._FIELD_MFA_METHODS:
            for method_name, method_config in value.items():
                if self._FIELD_HANDLER not in method_config:
                    raise MethodHandlerMissingError(method_name=method_name)
                for k, v in self.defaults[self._FIELD_MFA_METHODS][method_name].items():
                    method_config[k] = method_config.get(k, v)
                method_config[self._FIELD_HANDLER] = perform_import(
                    method_config[self._FIELD_HANDLER], self._FIELD_HANDLER
                )

    def __getitem__(self, attr: str) -> Any:
        return self.__getattr__(attr)


SOURCE_FIELD = "SOURCE_FIELD"
HANDLER = "HANDLER"
VALIDITY_PERIOD = "VALIDITY_PERIOD"
VERBOSE_NAME = "VERBOSE_NAME"
EMAIL_SUBJECT = "EMAIL_SUBJECT"
EMAIL_PLAIN_TEMPLATE = "EMAIL_PLAIN_TEMPLATE"
EMAIL_HTML_TEMPLATE = "EMAIL_HTML_TEMPLATE"
SMSAPI_ACCESS_TOKEN = "SMSAPI_ACCESS_TOKEN"
SMSAPI_FROM_NUMBER = "SMSAPI_FROM_NUMBER"
TWILIO_VERIFIED_FROM_NUMBER = "TWILIO_VERIFIED_FROM_NUMBER"
YUBICLOUD_CLIENT_ID = "YUBICLOUD_CLIENT_ID"

DEFAULTS = {
    "USER_MFA_MODEL": "trench.MFAMethod",
    "USER_ACTIVE_FIELD": "is_active",
    "BACKUP_CODES_QUANTITY": 5,
    "BACKUP_CODES_LENGTH": 12,  # keep (quantity * length) under 200
    "BACKUP_CODES_CHARACTERS": (string.ascii_letters + string.digits),
    "SECRET_KEY_LENGTH": 32,
    "DEFAULT_VALIDITY_PERIOD": 30,
    "CONFIRM_DISABLE_WITH_CODE": False,
    "CONFIRM_BACKUP_CODES_REGENERATION_WITH_CODE": True,
    "ALLOW_BACKUP_CODES_REGENERATION": True,
    "ENCRYPT_BACKUP_CODES": True,
    "APPLICATION_ISSUER_NAME": "MyApplication",
    "MFA_METHODS": {
        "sms_twilio": {
            VERBOSE_NAME: _("sms_twilio"),
            VALIDITY_PERIOD: 30,
            HANDLER: "trench.backends.twilio.TwilioMessageDispatcher",
            SOURCE_FIELD: "phone_number",
            TWILIO_VERIFIED_FROM_NUMBER: "YOUR TWILIO REGISTERED NUMBER",
        },
        "sms_api": {
            VERBOSE_NAME: _("sms_api"),
            VALIDITY_PERIOD: 30,
            HANDLER: "trench.backends.sms_api.SMSAPIMessageDispatcher",
            SOURCE_FIELD: "phone_number",
            SMSAPI_ACCESS_TOKEN: "YOUR SMSAPI TOKEN",
            SMSAPI_FROM_NUMBER: "YOUR REGISTERED NUMBER",
        },
        "email": {
            VERBOSE_NAME: _("email"),
            VALIDITY_PERIOD: 30,
            HANDLER: "trench.backends.basic_mail.SendMailMessageDispatcher",
            SOURCE_FIELD: "email",
            EMAIL_SUBJECT: _("Your verification code"),
            EMAIL_PLAIN_TEMPLATE: "trench/backends/email/code.txt",
            EMAIL_HTML_TEMPLATE: "trench/backends/email/code.html",
        },
        "app": {
            VERBOSE_NAME: _("app"),
            VALIDITY_PERIOD: 30,
            "USES_THIRD_PARTY_CLIENT": True,
            HANDLER: "trench.backends.application.ApplicationMessageDispatcher",
        },
        "yubi": {
            VERBOSE_NAME: _("yubi"),
            HANDLER: "trench.backends.yubikey.YubiKeyMessageDispatcher",
            YUBICLOUD_CLIENT_ID: "YOUR KEY",
        },
    },
}

trench_settings = TrenchAPISettings(
    user_settings=None, defaults=DEFAULTS, import_strings=None
)



from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.utils.crypto import constant_time_compare, salted_hmac
from django.utils.http import base36_to_int, int_to_base36
from django.utils.translation import gettext_lazy as _

from datetime import datetime
from typing import List, Optional, Tuple, Type

from trench.models import MFAMethod
from trench.settings import VERBOSE_NAME, trench_settings


class UserTokenGenerator(PasswordResetTokenGenerator):
    """
    Custom token generator:
        - user pk in token
        - expires after 15 minutes
        - longer hash (40 instead of 20)
    """

    KEY_SALT = "django.contrib.auth.tokens.PasswordResetTokenGenerator"
    SECRET = settings.SECRET_KEY
    EXPIRY_TIME = 60 * 15

    def make_token(self, user: User) -> str:
        return self._make_token_with_timestamp(user, int(datetime.now().timestamp()))

    def check_token(self, user: User, token: str) -> Optional[User]:
        user_model = get_user_model()
        if not token:
            return None
        try:
            token = str(token)
            user_pk, ts_b36, token_hash = token.rsplit("-", 2)
            ts = base36_to_int(ts_b36)
            user = user_model._default_manager.get(pk=user_pk)
        except (ValueError, TypeError, user_model.DoesNotExist):
            return None

        if (datetime.now().timestamp() - ts) > self.EXPIRY_TIME:
            return None  # pragma: no cover

        if not constant_time_compare(self._make_token_with_timestamp(user, ts), token):
            return None  # pragma: no cover

        return user

    def _make_token_with_timestamp(self, user: User, timestamp: int, **kwargs) -> str:
        ts_b36 = int_to_base36(timestamp)
        token_hash = salted_hmac(
            self.KEY_SALT,
            self._make_hash_value(user, timestamp),
            secret=self.SECRET,
        ).hexdigest()
        return f"{user.pk}-{ts_b36}-{token_hash}"


user_token_generator = UserTokenGenerator()


def get_mfa_model() -> Type[MFAMethod]:
    return apps.get_model(trench_settings.USER_MFA_MODEL)


def available_method_choices() -> List[Tuple[str, str]]:
    return [
        (method_name, method_config.get(VERBOSE_NAME, _(method_name)))
        for method_name, method_config in trench_settings.MFA_METHODS.items()
    ]



from django.contrib.auth.hashers import make_password

from typing import Callable, Set, Type

from trench.command.generate_backup_codes import generate_backup_codes_command
from trench.exceptions import MFAMethodDoesNotExistError
from trench.models import MFAMethod
from trench.settings import trench_settings
from trench.utils import get_mfa_model


class RegenerateBackupCodesForMFAMethodCommand:
    def __init__(
        self,
        requires_encryption: bool,
        mfa_model: Type[MFAMethod],
        code_hasher: Callable,
        codes_generator: Callable,
    ) -> None:
        self._requires_encryption = requires_encryption
        self._mfa_model = mfa_model
        self._code_hasher = code_hasher
        self._codes_generator = codes_generator

    def execute(self, user_id: int, name: str) -> Set[str]:
        backup_codes = self._codes_generator()
        rows_affected = self._mfa_model.objects.filter(
            user_id=user_id, name=name
        ).update(
            _backup_codes=MFAMethod._BACKUP_CODES_DELIMITER.join(
                [
                    self._code_hasher(backup_code)
                    for backup_code in backup_codes
                ] if self._requires_encryption else backup_codes
            ),
        )

        if rows_affected < 1:
            raise MFAMethodDoesNotExistError()

        return backup_codes


regenerate_backup_codes_for_mfa_method_command = (
    RegenerateBackupCodesForMFAMethodCommand(
        requires_encryption=trench_settings.ENCRYPT_BACKUP_CODES,
        mfa_model=get_mfa_model(),
        code_hasher=make_password,
        codes_generator=generate_backup_codes_command,
    ).execute
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2019-01-11 14:03
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("trench", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="mfamethod",
            old_name="backup_codes",
            new_name="_backup_codes",
        ),
        migrations.AlterField(
            model_name="mfamethod",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="mfa_methods",
                to=settings.AUTH_USER_MODEL,
                verbose_name="user",
            ),
        ),
    ]



from django.contrib.auth.models import User
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _

from abc import ABC, abstractmethod
from rest_framework.generics import ListAPIView
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.status import (
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
)
from rest_framework.views import APIView

from trench.backends.provider import get_mfa_handler
from trench.command.activate_mfa_method import activate_mfa_method_command
from trench.command.authenticate_second_factor import authenticate_second_step_command
from trench.command.authenticate_user import authenticate_user_command
from trench.command.create_mfa_method import create_mfa_method_command
from trench.command.deactivate_mfa_method import deactivate_mfa_method_command
from trench.command.replace_mfa_method_backup_codes import (
    regenerate_backup_codes_for_mfa_method_command,
)
from trench.command.set_primary_mfa_method import set_primary_mfa_method_command
from trench.exceptions import MFAMethodDoesNotExistError, MFAValidationError
from trench.query.get_mfa_config_by_name import get_mfa_config_by_name_query
from trench.responses import ErrorResponse
from trench.serializers import (
    ChangePrimaryMethodValidator,
    CodeLoginSerializer,
    LoginSerializer,
    MFAMethodActivationConfirmationValidator,
    MFAMethodBackupCodesGenerationValidator,
    MFAMethodCodeSerializer,
    MFAMethodDeactivationValidator,
    UserMFAMethodSerializer,
    generate_model_serializer,
)
from trench.settings import SOURCE_FIELD, trench_settings
from trench.utils import available_method_choices, get_mfa_model, user_token_generator


class MFAStepMixin(APIView, ABC):
    permission_classes = (AllowAny,)

    @abstractmethod
    def _successful_authentication_response(self, user: User) -> Response:
        raise NotImplementedError


class MFAFirstStepMixin(MFAStepMixin, ABC):
    def post(self, request: Request) -> Response:
        serializer = LoginSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            user = authenticate_user_command(
                request=request,
                username=serializer.validated_data[User.USERNAME_FIELD],
                password=serializer.validated_data["password"],
            )
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)
        try:
            mfa_model = get_mfa_model()
            mfa_method = mfa_model.objects.get_primary_active(user_id=user.id)
            get_mfa_handler(mfa_method=mfa_method).dispatch_message()
            return Response(
                data={
                    "ephemeral_token": user_token_generator.make_token(user),
                    "method": mfa_method.name,
                }
            )
        except MFAMethodDoesNotExistError:
            return self._successful_authentication_response(user=user)


class MFASecondStepMixin(MFAStepMixin, ABC):
    def post(self, request: Request) -> Response:
        serializer = CodeLoginSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            user = authenticate_second_step_command(
                code=serializer.validated_data["code"],
                ephemeral_token=serializer.validated_data["ephemeral_token"],
            )
            return self._successful_authentication_response(user=user)
        except MFAValidationError as cause:
            return ErrorResponse(error=cause, status=HTTP_401_UNAUTHORIZED)


class MFAMethodActivationView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request, method: str) -> Response:
        try:
            source_field = get_mfa_config_by_name_query(name=method).get(SOURCE_FIELD)
        except MFAMethodDoesNotExistError as cause:
            return ErrorResponse(error=cause)
        if source_field is not None:
            serializer_class = generate_model_serializer(
                name="MFAMethodActivationValidator",
                model=request.user.__class__,
                fields=(source_field,),
            )
            serializer = serializer_class(data=request.data)
            serializer.is_valid(raise_exception=True)
        try:
            mfa = create_mfa_method_command(
                user_id=request.user.id,
                name=method,
            )
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)
        return get_mfa_handler(mfa_method=mfa).dispatch_message()


class MFAMethodConfirmActivationView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request, method: str) -> Response:
        serializer = MFAMethodActivationConfirmationValidator(
            mfa_method_name=method, user=request.user, data=request.data
        )
        if not serializer.is_valid():
            return Response(status=HTTP_400_BAD_REQUEST, data=serializer.errors)
        try:
            backup_codes = activate_mfa_method_command(
                user_id=request.user.id,
                name=method,
                code=serializer.validated_data["code"],
            )
            return Response({"backup_codes": backup_codes})
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)


class MFAMethodDeactivationView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request, method: str) -> Response:
        serializer = MFAMethodDeactivationValidator(
            mfa_method_name=method, user=request.user, data=request.data
        )
        if not serializer.is_valid():
            return Response(status=HTTP_400_BAD_REQUEST, data=serializer.errors)
        try:
            deactivate_mfa_method_command(
                mfa_method_name=method, user_id=request.user.id
            )
            return Response(status=HTTP_204_NO_CONTENT)
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)


class MFAMethodBackupCodesRegenerationView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request, method: str) -> Response:
        if not trench_settings.ALLOW_BACKUP_CODES_REGENERATION:
            return ErrorResponse(error=_("Backup codes regeneration is not allowed."))
        serializer = MFAMethodBackupCodesGenerationValidator(
            mfa_method_name=method, user=request.user, data=request.data
        )
        if not serializer.is_valid():
            return Response(status=HTTP_400_BAD_REQUEST, data=serializer.errors)
        try:
            backup_codes = regenerate_backup_codes_for_mfa_method_command(
                user_id=request.user.id,
                name=method,
            )
            return Response({"backup_codes": backup_codes})
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)


class MFAConfigView(APIView):
    permission_classes = (AllowAny,)

    @staticmethod
    def get(request: Request) -> Response:
        return Response(
            data={
                "methods": [
                    method_name
                    for method_name, method_verbose_name in available_method_choices()
                ],
                "confirm_disable_with_code": trench_settings.CONFIRM_DISABLE_WITH_CODE,  # noqa
                "confirm_regeneration_with_code": trench_settings.CONFIRM_BACKUP_CODES_REGENERATION_WITH_CODE,  # noqa
                "allow_backup_codes_regeneration": trench_settings.ALLOW_BACKUP_CODES_REGENERATION,  # noqa
            },
        )


class MFAListActiveUserMethodsView(ListAPIView):
    serializer_class = UserMFAMethodSerializer
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet:
        mfa_model = get_mfa_model()
        return mfa_model.objects.list_active(user_id=self.request.user.id)


class MFAMethodRequestCodeView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request) -> Response:
        serializer = MFAMethodCodeSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            method = serializer.validated_data.get("method")
            mfa_model = get_mfa_model()
            if method is None:
                method = mfa_model.objects.get_primary_active_name(
                    user_id=request.user.id
                )
            mfa = mfa_model.objects.get_by_name(user_id=request.user.id, name=method)
            return get_mfa_handler(mfa_method=mfa).dispatch_message()
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)


class MFAPrimaryMethodChangeView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request) -> Response:
        mfa_model = get_mfa_model()
        mfa_method_name = mfa_model.objects.get_primary_active_name(
            user_id=request.user.id
        )
        serializer = ChangePrimaryMethodValidator(
            user=request.user, mfa_method_name=mfa_method_name, data=request.data
        )
        serializer.is_valid(raise_exception=True)
        try:
            set_primary_mfa_method_command(
                user_id=request.user.id, name=serializer.validated_data["method"]
            )
            return Response(status=HTTP_204_NO_CONTENT)
        except MFAValidationError as cause:
            return ErrorResponse(error=cause)



from django.utils.crypto import get_random_string

from typing import Callable, Set

from trench.settings import trench_settings


class GenerateBackupCodesCommand:
    def __init__(self, random_string_generator: Callable) -> None:
        self._random_string_generator = random_string_generator

    def execute(
        self,
        quantity: int = trench_settings.BACKUP_CODES_QUANTITY,
        length: int = trench_settings.BACKUP_CODES_LENGTH,
        allowed_chars: str = trench_settings.BACKUP_CODES_CHARACTERS,
    ) -> Set[str]:
        """
        Generates random encrypted backup codes.

        :param quantity: How many codes should be generated
        :type quantity: int
        :param length: How long codes should be
        :type length: int
        :param allowed_chars: Characters to create backup codes from
        :type allowed_chars: str

        :returns: Encrypted backup codes
        :rtype: set[str]
        """
        return {
            self._random_string_generator(length, allowed_chars)
            for _ in range(quantity)
        }


generate_backup_codes_command = GenerateBackupCodesCommand(
    random_string_generator=get_random_string,
).execute



from django.contrib.auth.models import User

from typing import Type

from trench.backends.provider import get_mfa_handler
from trench.command.remove_backup_code import remove_backup_code_command
from trench.command.validate_backup_code import validate_backup_code_command
from trench.exceptions import InvalidCodeError, InvalidTokenError
from trench.models import MFAMethod
from trench.utils import get_mfa_model, user_token_generator


class AuthenticateSecondFactorCommand:
    def __init__(self, mfa_model: Type[MFAMethod]) -> None:
        self._mfa_model = mfa_model

    def execute(self, code: str, ephemeral_token: str) -> User:
        user = user_token_generator.check_token(user=None, token=ephemeral_token)
        if user is None:
            raise InvalidTokenError()
        self.is_authenticated(user_id=user.id, code=code)
        return user

    def is_authenticated(self, user_id: int, code: str) -> None:
        for auth_method in self._mfa_model.objects.list_active(user_id=user_id):
            validated_backup_code = validate_backup_code_command(
                value=code, backup_codes=auth_method.backup_codes
            )
            if get_mfa_handler(mfa_method=auth_method).validate_code(code=code):
                return
            if validated_backup_code:
                remove_backup_code_command(
                    user_id=auth_method.user_id, method_name=auth_method.name, code=code
                )
                return
        raise InvalidCodeError()


authenticate_second_step_command = AuthenticateSecondFactorCommand(
    mfa_model=get_mfa_model()
).execute



from rest_framework.response import Response
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_400_BAD_REQUEST,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from trench.exceptions import MFAValidationError


class DispatchResponse(Response):
    _FIELD_DETAILS = "details"


class SuccessfulDispatchResponse(DispatchResponse):
    def __init__(
        self, details: str, status: str = HTTP_200_OK, *args, **kwargs
    ) -> None:
        super().__init__(
            data={self._FIELD_DETAILS: details}, status=status, *args, **kwargs
        )


class FailedDispatchResponse(DispatchResponse):
    def __init__(
        self, details: str, status: str = HTTP_422_UNPROCESSABLE_ENTITY, *args, **kwargs
    ) -> None:
        super().__init__(
            data={self._FIELD_DETAILS: details}, status=status, *args, **kwargs
        )


class ErrorResponse(Response):
    _FIELD_ERROR = "error"

    def __init__(
        self,
        error: MFAValidationError,
        status: str = HTTP_400_BAD_REQUEST,
        *args,
        **kwargs
    ) -> None:
        super().__init__(
            data={self._FIELD_ERROR: str(error)}, status=status, *args, **kwargs
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-10-12 11:34
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="MFAMethod",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=255, verbose_name="name"),
                ),
                (
                    "secret",
                    models.CharField(max_length=20, verbose_name="secret"),
                ),
                (
                    "is_primary",
                    models.BooleanField(default=False, verbose_name="is primary"),
                ),
                (
                    "is_active",
                    models.BooleanField(default=False, verbose_name="is active"),
                ),
                (
                    "backup_codes",
                    models.CharField(
                        blank=True, max_length=255, verbose_name="backup codes"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="mfa",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "MFA Method",
                "verbose_name_plural": "MFA Methods",
            },
        ),
    ]



from django.utils.translation import gettext_lazy as _

import logging
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client

from trench.backends.base import AbstractMessageDispatcher
from trench.responses import (
    DispatchResponse,
    FailedDispatchResponse,
    SuccessfulDispatchResponse,
)
from trench.settings import TWILIO_VERIFIED_FROM_NUMBER


class TwilioMessageDispatcher(AbstractMessageDispatcher):
    _SMS_BODY = _("Your verification code is: ")
    _SUCCESS_DETAILS = _("SMS message with MFA code has been sent.")

    def dispatch_message(self) -> DispatchResponse:
        try:
            client = Client()
            client.messages.create(
                body=self._SMS_BODY + self.create_code(),
                to=self._to,
                from_=self._config.get(TWILIO_VERIFIED_FROM_NUMBER),
            )
            return SuccessfulDispatchResponse(details=self._SUCCESS_DETAILS)
        except TwilioRestException as cause:
            logging.error(cause, exc_info=True)
            return FailedDispatchResponse(details=cause.msg)



from django.contrib.auth import user_logged_in, user_logged_out

from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from trench.serializers import TokenSerializer
from trench.views import MFAFirstStepMixin, MFASecondStepMixin, MFAStepMixin, User


class MFAAuthTokenView(MFAStepMixin):
    def _successful_authentication_response(self, user: User) -> Response:
        token, _ = Token.objects.get_or_create(user=user)
        user_logged_in.send(sender=user.__class__, request=self.request, user=user)
        return Response(data=TokenSerializer(token).data)


class MFAFirstStepAuthTokenView(MFAAuthTokenView, MFAFirstStepMixin):
    pass


class MFASecondStepAuthTokenView(MFAAuthTokenView, MFASecondStepMixin):
    pass


class MFALogoutView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def post(request: Request) -> Response:
        Token.objects.filter(user=request.user).delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return Response(status=status.HTTP_204_NO_CONTENT)



from trench.backends.base import AbstractMessageDispatcher
from trench.models import MFAMethod
from trench.query.get_mfa_config_by_name import get_mfa_config_by_name_query
from trench.settings import HANDLER


def get_mfa_handler(mfa_method: MFAMethod) -> AbstractMessageDispatcher:
    conf = get_mfa_config_by_name_query(name=mfa_method.name)
    dispatcher = conf[HANDLER]
    return dispatcher(mfa_method=mfa_method, config=conf)



from pyotp import TOTP


class CreateOTPCommand:
    @staticmethod
    def execute(secret: str, interval: int) -> TOTP:
        return TOTP(secret, interval=interval)


create_otp_command = CreateOTPCommand.execute



from django.utils.translation import gettext_lazy as _

import logging
from smsapi.client import SmsApiPlClient
from smsapi.exception import SmsApiException

from trench.backends.base import AbstractMessageDispatcher
from trench.responses import (
    DispatchResponse,
    FailedDispatchResponse,
    SuccessfulDispatchResponse,
)
from trench.settings import SMSAPI_ACCESS_TOKEN, SMSAPI_FROM_NUMBER


class SMSAPIMessageDispatcher(AbstractMessageDispatcher):
    _SMS_BODY = _("Your verification code is: ")
    _SUCCESS_DETAILS = _("SMS message with MFA code has been sent.")

    def dispatch_message(self) -> DispatchResponse:
        try:
            client = SmsApiPlClient(access_token=self._config.get(SMSAPI_ACCESS_TOKEN))
            from_number = self._config.get(SMSAPI_FROM_NUMBER)
            kwargs = {"from_": from_number} if from_number else {}
            client.sms.send(
                message=self._SMS_BODY + self.create_code(),
                to=self._to,
                **kwargs,
            )
            return SuccessfulDispatchResponse(details=self._SUCCESS_DETAILS)
        except SmsApiException as cause:
            logging.error(cause, exc_info=True)
            return FailedDispatchResponse(details=cause.message)



from django.contrib.auth.hashers import check_password

from typing import Iterable, Optional

from trench.settings import TrenchAPISettings, trench_settings


class ValidateBackupCodeCommand:
    def __init__(self, settings: TrenchAPISettings) -> None:
        self._settings = settings

    def execute(self, value: str, backup_codes: Iterable) -> Optional[str]:
        if not self._settings.ENCRYPT_BACKUP_CODES:
            return value if value in backup_codes else None
        for backup_code in backup_codes:
            if check_password(value, backup_code):
                return backup_code
        return None


validate_backup_code_command = ValidateBackupCodeCommand(
    settings=trench_settings
).execute



from typing import Any, Dict

from trench.exceptions import MFAMethodDoesNotExistError
from trench.settings import TrenchAPISettings, trench_settings


class GetMFAConfigByNameQuery:
    def __init__(self, settings: TrenchAPISettings) -> None:
        self._settings = settings

    def execute(self, name: str) -> Dict[str, Any]:
        try:
            return self._settings.MFA_METHODS[name]
        except KeyError as cause:
            raise MFAMethodDoesNotExistError from cause


get_mfa_config_by_name_query = GetMFAConfigByNameQuery(settings=trench_settings).execute



from typing import Callable, Set, Type

from trench.backends.provider import get_mfa_handler
from trench.command.generate_backup_codes import generate_backup_codes_command
from trench.command.replace_mfa_method_backup_codes import (
    regenerate_backup_codes_for_mfa_method_command,
)
from trench.exceptions import MFAMethodDoesNotExistError
from trench.models import MFAMethod
from trench.utils import get_mfa_model


class ActivateMFAMethodCommand:
    def __init__(
        self, mfa_model: Type[MFAMethod], backup_codes_generator: Callable
    ) -> None:
        self._mfa_model = mfa_model
        self._backup_codes_generator = backup_codes_generator

    def execute(self, user_id: int, name: str, code: str) -> Set[str]:
        mfa = self._mfa_model.objects.get_by_name(user_id=user_id, name=name)

        get_mfa_handler(mfa).confirm_activation(code)

        rows_affected = self._mfa_model.objects.filter(
            user_id=user_id, name=name
        ).update(
            is_active=True,
            is_primary=not self._mfa_model.objects.primary_exists(user_id=user_id),
        )

        if rows_affected < 1:
            raise MFAMethodDoesNotExistError()

        backup_codes = regenerate_backup_codes_for_mfa_method_command(
            user_id=user_id,
            name=name,
        )

        return backup_codes


activate_mfa_method_command = ActivateMFAMethodCommand(
    mfa_model=get_mfa_model(),
    backup_codes_generator=generate_backup_codes_command,
).execute



from pyotp import random_base32
from typing import Callable

from trench.settings import TrenchAPISettings, trench_settings


class CreateSecretCommand:
    def __init__(self, generator: Callable, settings: TrenchAPISettings) -> None:
        self._generator = generator
        self._settings = settings

    def execute(self) -> str:
        return self._generator(length=self._settings.SECRET_KEY_LENGTH)


create_secret_command = CreateSecretCommand(
    generator=random_base32, settings=trench_settings
).execute



from django.contrib.auth import authenticate
from django.contrib.auth.models import User

from rest_framework.request import Request

from trench.exceptions import UnauthenticatedError


class AuthenticateUserCommand:
    @staticmethod
    def execute(request: Request, username: str, password: str) -> User:
        user = authenticate(
            request=request,
            username=username,
            password=password,
        )
        if user is None:
            raise UnauthenticatedError()
        return user


authenticate_user_command = AuthenticateUserCommand.execute



from django.contrib.auth.hashers import check_password

from typing import Any, Set, Type

from trench.exceptions import InvalidCodeError, MFAMethodDoesNotExistError
from trench.models import MFAMethod
from trench.settings import TrenchAPISettings, trench_settings
from trench.utils import get_mfa_model


class RemoveBackupCodeCommand:
    def __init__(self, mfa_model: Type[MFAMethod], settings: TrenchAPISettings) -> None:
        self._mfa_model = mfa_model
        self._settings = settings

    def execute(self, user_id: Any, method_name: str, code: str) -> None:
        serialized_codes = (
            self._mfa_model.objects.filter(user_id=user_id, name=method_name)
            .values_list("_backup_codes", flat=True)
            .first()
        )
        if serialized_codes is None:
            raise MFAMethodDoesNotExistError()
        codes = MFAMethod._BACKUP_CODES_DELIMITER.join(
            self._remove_code_from_set(
                backup_codes=set(
                    serialized_codes.split(MFAMethod._BACKUP_CODES_DELIMITER)
                ),
                code=code,
            )
        )
        self._mfa_model.objects.filter(user_id=user_id, name=method_name).update(
            _backup_codes=codes
        )

    def _remove_code_from_set(self, backup_codes: Set[str], code: str) -> Set[str]:
        if not self._settings.ENCRYPT_BACKUP_CODES:
            backup_codes.remove(code)
            return backup_codes
        for backup_code in backup_codes:
            if check_password(code, backup_code):
                backup_codes.remove(backup_code)
                return backup_codes
        raise InvalidCodeError()


remove_backup_code_command = RemoveBackupCodeCommand(
    mfa_model=get_mfa_model(),
    settings=trench_settings,
).execute



from django.db.transaction import atomic

from typing import Type

from trench.exceptions import MFAMethodDoesNotExistError, MFAPrimaryMethodInactiveError
from trench.models import MFAMethod
from trench.utils import get_mfa_model


class SetPrimaryMFAMethodCommand:
    def __init__(self, mfa_model: Type[MFAMethod]) -> None:
        self._mfa_model = mfa_model

    @atomic
    def execute(self, user_id: int, name: str) -> None:
        self._mfa_model.objects.filter(user_id=user_id, is_primary=True).update(
            is_primary=False
        )
        if not self._mfa_model.objects.is_active_by_name(user_id=user_id, name=name):
            raise MFAPrimaryMethodInactiveError()
        rows_affected = self._mfa_model.objects.filter(
            user_id=user_id, name=name
        ).update(is_primary=True)
        if rows_affected < 1:
            raise MFAMethodDoesNotExistError()


set_primary_mfa_method_command = SetPrimaryMFAMethodCommand(
    mfa_model=get_mfa_model()
).execute



from django.urls import path

from trench.views.authtoken import MFAFirstStepAuthTokenView, MFASecondStepAuthTokenView


urlpatterns = (
    path("login/", MFAFirstStepAuthTokenView.as_view(), name="generate-code-authtoken"),
    path(
        "login/code/",
        MFASecondStepAuthTokenView.as_view(),
        name="generate-token-authtoken",
    ),
)



from setuptools import find_packages, setup

from trench import __version__


setup(
    name="django-trench",
    version=__version__,
    packages=find_packages(exclude=("testproject", "testproject.*")),
    include_package_data=True,
    license="MIT License",
    description="REST Multi-factor authentication package for Django",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    url="https://github.com/merixstudio/django-trench",
    author="Merixstudio",
    author_email="trench@merixstudio.com",
    install_requires=[
        "pyotp>=2.6.0",
        "twilio>=6.56.0",
        "yubico-client>=1.13.0",
        "smsapi-client>=2.4.5",
    ],
    extras_require={
        'docs': [
            'sphinx >= 1.4',
            'sphinx_rtd_theme',
        ]
    },
    classifiers=[
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



from typing import Callable, Type

from trench.command.create_secret import create_secret_command
from trench.exceptions import MFAMethodAlreadyActiveError
from trench.models import MFAMethod
from trench.utils import get_mfa_model


class CreateMFAMethodCommand:
    def __init__(self, secret_generator: Callable, mfa_model: Type[MFAMethod]) -> None:
        self._mfa_model = mfa_model
        self._create_secret = secret_generator

    def execute(self, user_id: int, name: str) -> MFAMethod:
        mfa, created = self._mfa_model.objects.get_or_create(
            user_id=user_id,
            name=name,
            defaults={
                "secret": self._create_secret,
                "is_active": False,
            },
        )
        if not created and mfa.is_active:
            raise MFAMethodAlreadyActiveError()
        return mfa


create_mfa_method_command = CreateMFAMethodCommand(
    secret_generator=create_secret_command, mfa_model=get_mfa_model()
).execute



from django.conf import settings
from django.db.models import (
    CASCADE,
    BooleanField,
    CharField,
    ForeignKey,
    Manager,
    Model,
    QuerySet,
    TextField,
)
from django.utils.translation import gettext_lazy as _

from typing import Any, Iterable

from trench.exceptions import MFAMethodDoesNotExistError


class MFAUserMethodManager(Manager):
    def get_by_name(self, user_id: Any, name: str) -> "MFAMethod":
        try:
            return self.get(user_id=user_id, name=name)
        except self.model.DoesNotExist:
            raise MFAMethodDoesNotExistError()

    def get_primary_active(self, user_id: Any) -> "MFAMethod":
        try:
            return self.get(user_id=user_id, is_primary=True, is_active=True)
        except self.model.DoesNotExist:
            raise MFAMethodDoesNotExistError()

    def get_primary_active_name(self, user_id: Any) -> str:
        method_name = (
            self.filter(user_id=user_id, is_primary=True, is_active=True)
            .values_list("name", flat=True)
            .first()
        )
        if method_name is None:
            raise MFAMethodDoesNotExistError()
        return method_name

    def is_active_by_name(self, user_id: Any, name: str) -> bool:
        is_active = (
            self.filter(user_id=user_id, name=name)
            .values_list("is_active", flat=True)
            .first()
        )
        if is_active is None:
            raise MFAMethodDoesNotExistError()
        return is_active

    def list_active(self, user_id: Any) -> QuerySet:
        return self.filter(user_id=user_id, is_active=True)

    def primary_exists(self, user_id: Any) -> bool:
        return self.filter(user_id=user_id, is_primary=True).exists()


class MFAMethod(Model):
    _BACKUP_CODES_DELIMITER = "|"

    user = ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=CASCADE,
        verbose_name=_("user"),
        related_name="mfa_methods",
    )
    name = CharField(_("name"), max_length=255)
    secret = CharField(_("secret"), max_length=255)
    is_primary = BooleanField(_("is primary"), default=False)
    is_active = BooleanField(_("is active"), default=False)
    _backup_codes = TextField(_("backup codes"), blank=True)

    class Meta:
        verbose_name = _("MFA Method")
        verbose_name_plural = _("MFA Methods")

    objects = MFAUserMethodManager()

    def __str__(self) -> str:
        return f"{self.name} (User id: {self.user_id})"

    @property
    def backup_codes(self) -> Iterable[str]:
        return self._backup_codes.split(self._BACKUP_CODES_DELIMITER)

    @backup_codes.setter
    def backup_codes(self, codes: Iterable) -> None:
        self._backup_codes = self._BACKUP_CODES_DELIMITER.join(codes)



from rest_framework.response import Response
from rest_framework_simplejwt.tokens import RefreshToken

from trench.views import MFAFirstStepMixin, MFASecondStepMixin, MFAStepMixin, User


class MFAJWTView(MFAStepMixin):
    def _successful_authentication_response(self, user: User) -> Response:
        token = RefreshToken.for_user(user=user)
        return Response(data={"refresh": str(token), "access": str(token.access_token)})


class MFAFirstStepJWTView(MFAJWTView, MFAFirstStepMixin):
    pass


class MFASecondStepJWTView(MFAJWTView, MFASecondStepMixin):
    pass



from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import get_template
from django.utils.translation import gettext_lazy as _

import logging
from smtplib import SMTPException

from trench.backends.base import AbstractMessageDispatcher
from trench.responses import (
    DispatchResponse,
    FailedDispatchResponse,
    SuccessfulDispatchResponse,
)
from trench.settings import EMAIL_HTML_TEMPLATE, EMAIL_PLAIN_TEMPLATE, EMAIL_SUBJECT


class SendMailMessageDispatcher(AbstractMessageDispatcher):
    _KEY_MESSAGE = "message"
    _SUCCESS_DETAILS = _("Email message with MFA code has been sent.")

    def dispatch_message(self) -> DispatchResponse:
        context = {"code": self.create_code()}
        email_plain_template = self._config[EMAIL_PLAIN_TEMPLATE]
        email_html_template = self._config[EMAIL_HTML_TEMPLATE]
        try:
            send_mail(
                subject=self._config.get(EMAIL_SUBJECT),
                message=get_template(email_plain_template).render(context),
                html_message=get_template(email_html_template).render(context),
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=(self._to,),
                fail_silently=False,
            )
            return SuccessfulDispatchResponse(details=self._SUCCESS_DETAILS)
        except SMTPException as cause:  # pragma: nocover
            logging.error(cause, exc_info=True)  # pragma: nocover
            return FailedDispatchResponse(details=str(cause))  # pragma: nocover



from django.urls import path


__all__ = [
    "urlpatterns",
]

from trench.views import (
    MFAConfigView,
    MFAListActiveUserMethodsView,
    MFAMethodActivationView,
    MFAMethodBackupCodesRegenerationView,
    MFAMethodConfirmActivationView,
    MFAMethodDeactivationView,
    MFAMethodRequestCodeView,
    MFAPrimaryMethodChangeView,
)


urlpatterns = (
    path(
        "<str:method>/activate/",
        MFAMethodActivationView.as_view(),
        name="mfa-activate",
    ),
    path(
        "<str:method>/activate/confirm/",
        MFAMethodConfirmActivationView.as_view(),
        name="mfa-activate-confirm",
    ),
    path(
        "<str:method>/deactivate/",
        MFAMethodDeactivationView.as_view(),
        name="mfa-deactivate",
    ),
    path(
        "<str:method>/codes/regenerate/",
        MFAMethodBackupCodesRegenerationView.as_view(),
        name="mfa-regenerate-codes",
    ),
    path("code/request/", MFAMethodRequestCodeView.as_view(), name="mfa-request-code"),
    path("mfa/config/", MFAConfigView.as_view(), name="mfa-config-info"),
    path(
        "mfa/user-active-methods/",
        MFAListActiveUserMethodsView.as_view(),
        name="mfa-list-user-active-methods",
    ),
    path(
        "mfa/change-primary-method/",
        MFAPrimaryMethodChangeView.as_view(),
        name="mfa-change-primary-method",
    ),
)
